__all__ = [
    'AppHandle',
    'Application',
    'Config',
    'LoggingModule',
    'Module',
]

from applipy.version import __version__  # noqa
from applipy.application.apphandle import AppHandle
from applipy.application.application import Application
from applipy.application.module import Module
from applipy.config.config import Config
from applipy.logging.module import LoggingModule



import signal
import sys
import traceback

from typing import Union, Callable, Type, Optional, TypeVar, Any
from asyncio import gather, get_event_loop, BaseEventLoop, CancelledError
from logging import Logger, INFO, ERROR, getLevelName
from applipy.application.apphandle import AppHandle
from applipy.application.module import Module
from applipy.config.config import Config
from applipy_inject import Injector

if sys.version_info[:2] <= (3, 6):
    from asyncio import Task
    all_tasks = Task.all_tasks
else:
    from asyncio import all_tasks as all_tasks


T = TypeVar('T')
BindFunction = Callable[
    [
        Type[T],
        Union[T, Callable[[Any], T]],
        Optional[str],
        Optional[bool]
    ],
    None
]
RegisterFunction = Callable[[Module], None]


class ModuleManager:

    def __init__(self, injector: Injector) -> None:
        self._injector = injector
        self._modules = set()

    def install(self, module: Type[Module]) -> 'ModuleManager':
        if module not in self._modules:
            self._modules.add(module)
            self._injector.bind_provider(Module, module)
            for dep in module.depends_on():
                self.install(dep)

        return self

    def configure_all(self, bind_function: BindFunction,
                      register_function: RegisterFunction) -> None:
        instances = self.injector.get_all(Module)
        for instance in instances:
            instance.configure(bind_function, register_function)
            self._log(INFO, f'Installing module `{instance.__class__.__module__}.{instance.__class__.__name__}`')

        self._modules = set()

    @property
    def injector(self):
        return self._injector

    def _log(self, level, *args, **kwargs):
        try:
            self._injector.get(Logger).log(level, *args, **kwargs)
        except Exception:
            print(f"[{getLevelName(level)}]", *args)


class EventLoopManager:

    def __init__(self, loop=None):
        self._loop = loop or get_event_loop()

    def run(self, coro):
        self._loop.run_until_complete(coro)

    def shutdown(self):
        self.cancel_loop_tasks()
        self._close_loop()

    def _close_loop(self):
        self._loop.run_until_complete(self._loop.shutdown_asyncgens())
        self._loop.close()

    def cancel_loop_tasks(self):
        tasks_to_cancel = {t for t in all_tasks(self._loop)
                           if not t.done()}
        if not tasks_to_cancel:
            return

        for t in tasks_to_cancel:
            t.cancel()


AppHandleProvider = Union[Type[AppHandle], Callable[[], AppHandle]]


class AppHandleManager:

    def __init__(self, injector: Injector, loop: BaseEventLoop) -> None:
        self._injector = injector
        self._loop = loop

    def register(self,
                 app_handle_provider: AppHandleProvider) -> 'AppHandleManager':
        self._injector.bind_provider(AppHandle, app_handle_provider)

    async def init_all(self):
        app_handles = self._injector.get_all(AppHandle)
        await gather(*(app_handle.on_init() for app_handle in app_handles),
                     loop=self._loop)

    async def start_all(self):
        app_handles = self._injector.get_all(AppHandle)
        await gather(*(app_handle.on_start() for app_handle in app_handles),
                     loop=self._loop)

    async def shutdown_all(self):
        app_handles = self._injector.get_all(AppHandle)
        await gather(*(app_handle.on_shutdown() for app_handle in app_handles),
                     loop=self._loop)


class Application:

    def __init__(self,
                 config,
                 injector=None,
                 event_loop_manager=None,
                 module_manager=None,
                 app_handle_manager=None,
                 loop=None,
                 signal_=None):
        self._config = config

        loop = loop or get_event_loop()
        self._injector = injector or Injector()

        self._event_loop_manager = event_loop_manager or EventLoopManager(loop)
        self._app_handle_manager = (app_handle_manager or
                                    AppHandleManager(self._injector, loop))

        self._module_manager = module_manager or ModuleManager(self._injector)

        def _sigterm_handler():
            self._log(INFO, "Received SIGTERM. Shutting down.")
            self._event_loop_manager.cancel_loop_tasks()

        def _sigint_handler():
            self._log(INFO, "Received SIGINT. Shutting down.")
            self._event_loop_manager.cancel_loop_tasks()

        loop.add_signal_handler(signal.SIGTERM, _sigterm_handler)
        loop.add_signal_handler(signal.SIGINT, _sigint_handler)

    def install(self, module):
        self._module_manager.install(module)
        return self

    def register(self, app_handle_provider):
        self._app_handle_manager.register(app_handle_provider)
        return self

    @property
    def injector(self):
        return self._injector

    def run(self):
        self._injector.bind(Config, self._config)
        self._module_manager.configure_all(
            self._injector.bind,
            self._app_handle_manager.register
        )
        self._event_loop_manager.run(self._run_application())
        self._event_loop_manager.shutdown()

    async def _run_application(self):
        try:
            await self._app_handle_manager.init_all()
            await self._app_handle_manager.start_all()
        except CancelledError:
            pass
        except Exception:
            self._log(ERROR, traceback.format_exc())
        finally:
            await self._app_handle_manager.shutdown_all()

    def _log(self, level, *args, **kwargs):
        try:
            self._injector.get(Logger).log(level, *args, **kwargs)
        except Exception:
            print(f"[{getLevelName(level)}]", *args)



import logging
from typing import List

from applipy.application.module import Module
from applipy.config.config import Config


class LoggingModule(Module):

    def __init__(self, config: Config):
        self.config = config

    def configure(self, bind, register):
        logging.basicConfig(**self.config.get('logging.config', {}))
        logger = logging.getLogger(self.config.get('app.name'))

        logging_level = self.config.get('logging.level')
        if logging_level:
            logger.setLevel(logging_level)

        def configure_logger(
                handlers: List[logging.Handler]
        ) -> logging.Logger:
            for handler in handlers:
                logger.addHandler(handler)

            return logger

        bind(configure_logger)



import sys
import os
import time
from subprocess import Popen


class ApplipyProcess:

    def __init__(self, config_path, applipy_env, env={}):
        self._env = os.environ.copy()
        self._env['APPLIPY_CONFIG_PATH'] = config_path
        self._env['APPLIPY_ENV'] = applipy_env
        self._env.update(env)

    def __enter__(self):
        self._process = Popen(('python', '-m', 'applipy'), stdout=sys.stdout, stderr=sys.stderr, env=self._env)
        time.sleep(1)
        return self._process

    def __exit__(self, *args):
        self._process.terminate()
        self._process.wait()
        return False



import os
import json
import sys
from pydoc import locate
from applipy import Application, Config
from applipy_inject import Injector
from applipy.application.application import (
    BindFunction,
    ModuleManager,
    RegisterFunction,
)
from logging import Logger, getLevelName, ERROR

try:
    import yaml
except ImportError:
    yaml = None


class LoadFromConfigModuleManager(ModuleManager):

    def set_application(self, app: Application) -> None:
        self.app = app

    def configure_all(self, bind_function: BindFunction,
                      register_function: RegisterFunction) -> None:
        config = self.injector.get(Config)

        module_names = config.get('app.modules')
        if module_names:
            for module, name in ((locate(name), name) for name in module_names):
                if module:
                    self.app.install(module)
                else:
                    self._log(ERROR, f'Could not load module `{name}`')
                    raise ImportError(name)

        super().configure_all(bind_function, register_function)

    def _log(self, level, *args, **kwargs):
        try:
            self.injector.get(Logger).log(level, *args, **kwargs)
        except Exception:
            print(f"[{getLevelName(level)}]", *args)


def _is_file(path):
    return os.path.isfile(path) or (os.path.exists(path) and not os.path.isdir(path))


def load_config_from_json(config_file):
    if _is_file(config_file):
        with open(config_file, 'r') as f:
            config = json.load(f)
    else:
        config = {}
    return config


def load_config_from_yaml(config_file):
    is_file = _is_file(config_file)
    if yaml and is_file:
        with open(config_file, 'r') as f:
            config = yaml.load(f, Loader=yaml.Loader)
    else:
        if is_file:
            print('[WARN] Found yaml file but no yaml module installed')
        config = {}
    return config


def load_raw_config(config_path, env):
    return {
        **load_config_from_yaml(os.path.join(config_path, f'{env.lower()}.yaml')),
        **load_config_from_json(os.path.join(config_path, f'{env.lower()}.json')),
    }


def build_config(config_raw):
    config = Config(config_raw)

    provider_names = config.get('config.protocols')
    if provider_names:
        for provider, name in ((locate(name), name) for name in provider_names):
            if provider:
                print(f'[INFO] Adding configuration provider `{name}`')
                config.addProtocol(provider())
            else:
                print(f'[ERROR] Could not load configuration provider `{name}`')
                raise ImportError(name)

    return config


def start(config):
    injector = Injector()
    module_manager = LoadFromConfigModuleManager(injector)
    app = Application(config, injector=injector, module_manager=module_manager)
    module_manager.set_application(app)
    app.run()


def main(config_path, env):
    start(build_config(load_raw_config(config_path, env)))


def main_cmd_yaml(config_file):
    config = load_config_from_yaml(config_file)
    start(build_config(config))


def entrypoint():
    config_file = None
    try:
        idx = sys.argv.index('-f')
        config_file = sys.argv[idx + 1]
    except IndexError:
        config_file = '/dev/stdin'
    except ValueError:
        ...

    if config_file:
        main_cmd_yaml(config_file)
    else:
        main(os.environ.get('APPLIPY_CONFIG_PATH', os.path.curdir), os.environ.get('APPLIPY_ENV', 'DEV'))


if __name__ == '__main__':
    entrypoint()



from setuptools import setup, find_packages
from distutils.util import convert_path
from os import path

ns = {}
ver_path = convert_path('applipy/version.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), ns)
version = ns['__version__']

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='applipy',
    url='https://gitlab.com/applipy/applipy',
    project_urls={
        'Source': 'https://gitlab.com/applipy/applipy',
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries',
    ],
    description='Library for building modular applications',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Apache 2.0',
    author='Alessio Linares',
    author_email='mail@alessio.cc',
    version=version,
    entry_points={
        'console_scripts': [
            'applipy = applipy.__main__:entrypoint',
        ],
    },
    packages=find_packages(exclude=['docs', 'tests']),
    data_files=[],
    python_requires='>=3.6',
    install_requires=['applipy_inject>=1.0.0,<2.0.0'],
    scripts=[],
)



from typing import Optional
from applipy.config.protocol import ConfigProtocol


class DockerSecrets(ConfigProtocol):

    def provide_for(self, protocol: str, key: str) -> Optional[str]:
        if protocol != 'secret':
            return None

        with open('/run/secrets/' + key, 'r') as f:
            return f.read()



from applipy import Config
from applipy.config.protocol import ConfigProtocol


class Upper(ConfigProtocol):

    def provide_for(self, protocol, key):
        if protocol in ('up', 'upper'):
            return key.upper()

        return None


class Capitalize(ConfigProtocol):

    def provide_for(self, protocol, key):
        if protocol == 'cap':
            return key.capitalize()

        return None


def test_config():

    raw_config = {
            'app': {'name': 'config-test', 'priority': 40},
            'app.type': 5,
            'db.password.value': 'up:seCret',
            'db.password.is_secret': True,
            'name': {'of': {'process': 'cap:foo'}},
            'name.of.file': 'upper:unknown',
            'private.files': [{'name': 'foo', 'meta': {'type': 'json'}}, {'name': 'bar', 'meta.type': 'yaml'}],
            'list.of.lists': [
                [{'a': 1}, {'b': 2}],
                [{'c': {'d': 3}}],
            ],
            'indirect.list.recursion': [{'foo': [{'bar': [{'z': {'w': 'upper:waves'}}]}]}]
    }
    config = Config(raw_config, [Capitalize()])
    config.addProtocol(Upper())

    assert config.keys() == {'app.name', 'app.priority', 'app.type', 'db.password.value', 'db.password.is_secret',
                             'name.of.process', 'name.of.file', 'private.files', 'list.of.lists',
                             'indirect.list.recursion'}
    assert config.get('private.files')[0].keys() == {'name', 'meta.type'}
    assert config.get('private.files')[1].keys() == {'name', 'meta.type'}

    assert config['app.name'] == 'config-test'
    assert config.get('app.priority') == 40
    assert config.get('app.type') == 5
    assert config.get('db.password.value') == 'SECRET'
    assert config.get('db.password.is_secret') is True
    assert config['name'].get('of.process') == 'Foo'
    assert config['name']['of.file'] == 'UNKNOWN'
    assert config.get('app.not-there', 69) == 69
    assert config['private.files'][0].get('name') == 'foo'
    assert config['private.files'][0]['meta.type'] == 'json'
    assert config['private.files'][1]['meta']['type'] == 'yaml'
    assert config['list.of.lists'][1][0]['c.d'] == 3
    assert config['indirect.list.recursion'][0]['foo'][0]['bar'][0]['z.w'] == 'WAVES'


def test_protocols():

    upper = Upper()
    capitalize = Capitalize()
    config = Config({}, [upper])

    assert len(config.protocols) == 1
    assert config.protocols[0] is upper

    config.addProtocol(capitalize)

    assert len(config.protocols) == 2
    assert config.protocols[0] is upper
    assert config.protocols[1] is capitalize
